-- toLua: accessing C/C++ code from Lua
-- Waldemar Celes
-- Mar 1997


$debug

-- type table: match C type with Lua type.
-- if a type does not have corresponding Lua type,
-- it is considered a generic object.
TYPE = {} 
TYPE[""]           = ""		-- because it's not a generic type
TYPE["void"]       = ""		-- because it's not a generic type
TYPE["char"]       = "number"
TYPE["short"]      = "number"
TYPE["int"]        = "number"
TYPE["long"]       = "number"
TYPE["float"]      = "number"
TYPE["double"]     = "number"
TYPE["char*"]      = "string"
TYPE["void*"]      = "userdata"
TYPE["lua_Object"] = "object"

-- tags
TAG = {}
TAG[""]           = ""
TAG["void"]       = ""
TAG["char"]       = "TOLUA_NUMBER_TAG"
TAG["short"]      = "TOLUA_NUMBER_TAG"
TAG["int"]        = "TOLUA_NUMBER_TAG"
TAG["long"]       = "TOLUA_NUMBER_TAG"
TAG["float"]      = "TOLUA_NUMBER_TAG"
TAG["double"]     = "TOLUA_NUMBER_TAG"
TAG["char*"]      = "TOLUA_STRING_TAG"
TAG["void*"]      = "TOLUA_USERDATA_TAG"
TAG["lua_Object"] = ""               -- doesn't matter

-- valid operator and their corresponding name suffixs in Lua
OPERATOR = {}
OPERATOR["[]"] = "index"
OPERATOR["+"]  = "add"
OPERATOR["-"]  = "sub"
OPERATOR["*"]  = "mult"
OPERATOR["/"]  = "div"
OPERATOR["<"]  = "lt"
OPERATOR[">"]  = "gt"
OPERATOR["<="] = "le"
OPERATOR[">="] = "ge"

-- Bind variables
-- Write C/C++ code to bind variables
-- As input it receives a variable list and an optimal class name
function bindVariables (v,class)
 local i=1
 while v[i] do
  -- compute function C name
  local n = "toLua_"
  if class then
   n = n..class.."_"
  else
   n = n..package.."_"
  end
  v[i].cname = n..v[i].name

  -- compute varname
  local varname = v[i].name
  if class then
--   if strfind(v[i].modifiers,"static") then
--    varname = class.."::"..varname
--   else
    varname = "self->"..varname
--   end
  end

  -- bind type tags
  bindTypeTag(v[i].type)
  if class then
    bindTypeTag(class)
  end

  -- write function to get
  if not class then
   write("#ifdef __cplusplus\n")
   write("static void "..v[i].cname.."_get(void*)\n")
   write("#else\n")
   write("static void "..v[i].cname.."_get(void* dummy)\n")
   write("#endif\n")
  else
   write("static void "..v[i].cname.."_get("..class.."* self)\n")
  end
  write("{\n ") 
  write(v[i].type..v[i].pointer.." L_result = ("..v[i].type..v[i].pointer..")")
  write(varname)
  write(";\n")
  if TYPE[v[i].type] then
   write(" toLua_push"..TYPE[v[i].type].."(L_result);\n")
  elseif v[i].pointer ~= "" then
   write(" toLua_pushusertag(")
   if v[i].pointer == "*" then
    write("(void*)L_result")
   elseif v[i].pointer == "&" then
    write("(void*)&L_result")
   else
    error("toLua - invalid pointer.")
   end
   write(","..TAG[v[i].type]..");\n")
  else
   write(" toLua_pushuservalue((void*)&L_result,sizeof(L_result),")
   write(TAG[v[i].type]..");\n")
  end
  write("}\n\n") 
 
  -- write function to set, if not const
  if not strfind(v[i].modifiers,"const") then
   if not class then
    write("#ifdef __cplusplus\n")
    write("static void "..v[i].cname.."_set(void*, lua_Object lo)\n")
    write("#else\n")
    write("static void "..v[i].cname.."_set(void* dummy, lua_Object lo)\n")
    write("#endif\n")
   else
    write("static void "..v[i].cname.."_set("..class.."* self, lua_Object lo)\n")
   end
   write("{\n") 
 
   -- eliminate "extern" and "static" modifiers
   local modifiers = gsub(v[i].modifiers,"extern%s*","")
   local modifiers = gsub(modifiers,"static%s*","")
 
   if v[i].type == "lua_Object" then
    write(" if (lo == LUA_NOOBJECT)\n")
   else
    write(" if (!toLua_istype(lo,"..TAG[v[i].type].."))\n")
   end
   write("  lua_error(\"tolua: incompatible variable type.\");\n")
   write(" else\n")
   write("  "..varname.." = ")
   if not TYPE[v[i].type] and v[i].pointer ~= "*" then
    write("*")
   end
   write("(")
   if modifiers ~= "" then
    write(modifiers.." ")
   end
   write(v[i].type)
   if not TYPE[v[i].type] then
    write("*")
   end
   write(")")
   if TYPE[v[i].type] then
    write("lua_get"..TYPE[v[i].type].."(lo);\n")
   else
    write("toLua_getusertag(lo,"..TAG[v[i].type]..");\n")
   end
   write("}\n\n") 
  end
 
  -- write function to initialize variable bind
  if not class then
   write("static void "..v[i].cname.."_init (void)\n{\n")
   write(" toLua_getfb_extern(\""..package.."\",\""..v[i].name.."\","..
         v[i].cname.."_get);\n"
        ) 
   if not strfind(v[i].modifiers,"const") then
    write(" toLua_setfb_extern(\""..package.."\",\""..v[i].name.."\","..
          v[i].cname.."_set);\n"
         ) 
   end
   write("}\n\n\n")
  else
   write("static void "..v[i].cname.."_init (void)\n{\n")
   write(" toLua_getfb("..TAG[class]..",\""..v[i].name.."\","..
         v[i].cname.."_get);\n"
        )
   if not strfind(v[i].modifiers,"const") then
    write(" toLua_setfb("..TAG[class]..",\""..v[i].name.."\","..
          v[i].cname.."_set);\n"
         )
   end
   write("}\n\n\n")
  end
  v[i].cname = v[i].cname.."_init"
  i = i+1
 end
end
 
-- Bind functions
-- Write C/C++ code to bind functions
-- As input it receives a function list and an optimal class name
function bindFunctions (t,class)

 -- compute function C names
 local i=1
 while t[i] do
  if i==1 or t[i].name ~= t[i-1].name then
   -- not overloaded names
   local n = "toLua"
   local fn = t[i].name
   -- deal with operators
   if strfind(t[i].name,"^operator") then
    local b,e,op = strfind(t[i].name,"^operator(.*)")
    if not OPERATOR[op] then
     error("Invalid 'operator'.")
    else
     fn = "operator_" .. OPERATOR[op]
    end
   end
   -- deal with class 
   if class then 
    n = n.."_"..class
    if class == t[i].name then
     n = n
    elseif "~"..class == t[i].name then
     n = n.."_delete"
    else
     n = n.."_"..fn
    end
   else
    n = n.."_"..fn
   end
   t[i].cname = n
  else
   -- overloaded names
   t[i].cname = t[i-1].cname.."_"..t[i].name
  end
  i = i+1
 end

 -- compute Lua names
 local i=1
 while t[i] do
  t[i].luaname = t[i].name 	-- that's the default

  -- deal with operators
  if strfind(t[i].name,"^operator") then
   local b,e,op = strfind(t[i].name,"^operator(.*)")
   if not OPERATOR[op] then
    error("Invalid 'operator'.")
   else
    t[i].luaname = "operator_" .. OPERATOR[op]
   end
  end

  -- deal with special class methods
  if class then
   if class == t[i].name then
    t[i].luaname = "new"
    t[i].type = class
    t[i].pointer = "*" 
   elseif "~"..class == t[i].name then 
    t[i].luaname = "delete"
    t[i].type = "void"
   end
  end
  i = i+1
 end

 -- bind each function
 local i=1
 while t[i] do
  bindOneFunction (t[i],t[i-1],class)
  i = i+1
 end
end

-- create tag, if necessary
function bindTypeTag (type)
 if not TAG[type] then
  TAG[type] = "L_tag_"..type
  TAG[type] = gsub(TAG[type],"[<>]","_")
  write("static int "..TAG[type]..";\n")
  -- record created tag
  USERTAG.n = USERTAG.n + 1
  USERTAG[USERTAG.n] = {tag = TAG[type], type = type}
 end
end

-- write C code to bind one function
function bindOneFunction (f,fprev,class)

 -- deal with parameter without name
 local i=1
 while f.args[i] do 
   local name = f.args[i].name
   if strsub(name,strlen(name)) == "*" then
    name = strsub(name,1,strlen(name)-1)
   elseif strsub(name,strlen(name)) == "&" then
    name = strsub(name,1,strlen(name)-1)
   end
   if name ~= "" and name ~= "void" and TAG[name] then
    f.args[i].modifiers,f.args[i].type,f.args[i].pointer,dummy, 
    f.args[i].name,dummy,f.args[i].dimension =  
     parseFullName(f.args[i].modifiers.." "..f.args[i].type.." "..
                   f.args[i].name.." _par"..i, 1
                  )
  end
  i = i+1
 end

 -- bind "self" type, if class
 if class then
  bindTypeTag(class)
 end

 -- bind type tags
 bindTypeTag(f.type)
 local i=1
 while f.args[i] do
  bindTypeTag(f.args[i].type)
  i = i+1
 end

 -- take care of operator[]
 if strfind(f.cname,"operator_index") then
  if not class then
   error("toLua - operator[] can only be defined within a class.")
  end
  bindOperatorIndex(f,class)
  return
 end

 -- open function
 write("static void "..f.cname.." (void)\n{\n") 

 -- declare "self" parameter, if class and if necessary
 if class and f.name ~= class and not strfind(f.modifiers,"static") then
  write(" "..class.."* self;\n")
 end

 -- declare parameters
 if f.args[1] ~= "" and f.args[1].type ~= "void" then 
  local i=1
  while f.args[i] do
   write(" ");
   if f.args[i].modifiers ~= "" then
    write(f.args[i].modifiers) write(" ")
   end
   write(f.args[i].type) 
   if not TYPE[f.args[i].type] then
    write("*")
   end
   write(" ")
   write(f.args[i].name) 
   if f.args[i].dimension ~= "" then
    if not TYPE[f.args[i].type] and f.args[i].pointer ~= "*" then
     error("only vector of basic types or vector of pointers are allowed.")
    end
    write("["..f.args[i].dimension.."]")
   end
   write(";\n")
   i = i+1
  end
 end

 -- check and get "self" parameter, if class
 if class and f.name ~= class and not strfind(f.modifiers,"static") then
  write("\n")
  write(" if (!toLua_nistype(1,"..TAG[class].."))")
  write(" {toLua_wrongself(\""..f.name.."\"); return;}\n")
  write(" else")
  write(" self = ("..class.."*)toLua_ngetusertag(1,"..TAG[class]..");\n")
 end
 
 -- check and get parameter values
 local i=1
 local j=1
 if class then j = j+1 end
 if f.args[i] ~= "" and f.args[i].type ~= "void" then
  while f.args[i] do
   write("\n")
   if f.args[i].dimension ~= "" then
    write(" if (!toLua_nisarray("..j..","..TAG[f.args[i].type]..","..
                f.args[i].dimension.."))"
         )
   elseif f.args[i].type == "lua_Object" then
    write(" if (lua_getparam("..j..") == LUA_NOOBJECT)")
   else
    write(" if (!toLua_nistype("..j..","..TAG[f.args[i].type].."))")
   end
   -- cannot have default value to vectors
   if f.args[i].default ~= "" and f.args[i].dimension == "" then
    write("\n {\n")
    write("  if (lua_getparam("..j..") == LUA_NOOBJECT)")
    write(" "..f.args[i].name.." = "..f.args[i].default..";\n")
    write("  else")
    if fprev and fprev.name == f.name then
     write(" { "..fprev.cname.."(); return; }\n")
    else
     write(" {toLua_wrongarg(\""..f.name.."\","..i..",\""..
           f.args[i].type.."\"); return;}\n") 
    end
    write(" }\n")
   elseif fprev and fprev.name == f.name then
    write(" { "..fprev.cname.."(); return; }\n")
   else
    if f.args[i].dimension == "" then
     write(" {toLua_wrongarg(\""..f.name.."\","..i..",\""..
           f.args[i].type.."\"); return;}\n")
    else
     write(" {toLua_wrongarray(\""..f.name.."\","..i..",\""..
           f.args[i].type.."\","..f.args[i].dimension.."); return;}\n")
    end
   end
   write(" else")
   local getfunc
   if f.args[i].dimension == "" then
    getprefix = "toLua_nget"
    indexname = j
   else
    getprefix = "toLua_iget"
    indexname = "i+1"
   end
   if f.args[i].dimension == "" then
    write("  "..f.args[i].name.." = (")
   else
    write(" {\n")
    write("  int i;\n")
    write("  lua_Object lt = toLua_ngetobject("..j..");\n")
    write("  lua_beginblock();\n");
    write("  for(i=0; i<"..f.args[i].dimension..";i++)\n")
    write("   "..f.args[i].name.."[i] = (")
   end
   if f.args[i].modifiers ~= "" then
    write(f.args[i].modifiers.." ")
   end
   write(f.args[i].type)
   if not TYPE[f.args[i].type] then
    write("*")
   end
   write(")")
   if TYPE[f.args[i].type] then
    write(getprefix..TYPE[f.args[i].type].."("..indexname)
   else
    write(getprefix.."usertag".."("..indexname..","..TAG[f.args[i].type])
   end 
   if f.args[i].dimension == "" then
    write(");\n")
   else
    write(",lt);\n")
    write("  lua_endblock();\n");
    write(" }\n")
   end
   i = i+1
   j = j+1
  end
 end

 -- check end of parameter list
 write("\n")
 write( " if (lua_getparam("..j..") != LUA_NOOBJECT)")
 if  fprev and fprev.name == f.name then
   write("{ "..fprev.cname.."(); return; }\n")
 else
   write(" toLua_toomanyargs(\""..f.name.."\");\n")
 end

 -- call function
 write("\n {\n  ")
 if f.type ~= "void" then
  write(f.type..f.pointer.." L_result = ("..f.type..f.pointer..")")
 end
 if class and f.luaname == "new" then
   write("new "..class.."(")
   writeParamList(f.args)
   write(");\n")
 elseif class and f.luaname == "delete" then
  write("delete self;\n")
 else
  if class then
   if strfind(f.modifiers,"static") then
    write(class.."::")
   else
    write("self->")
    if f.domain ~= "" then
     write(f.domain.."::")
    end
   end 
  end
  write(f.name.."(")
  writeParamList(f.args)
  write(");\n")
 end

 -- fill Lua vectors back
 i=1; local j=1
 if class then j = j+1 end
 if f.args[i] ~= "" and f.args[i].type ~= "void" then
  while f.args[i] do
   if f.args[i].dimension ~= "" then
    write("  {\n")
    write("   int i;\n")
    write("   lua_Object lt = toLua_ngetobject("..j..");\n")
    write("   lua_beginblock();\n");
    write("   for(i=0; i<"..f.args[i].dimension..";i++)\n")
    write("   {\n");
    write("    lua_pushobject(lt);\n");
    write("    toLua_pushnumber(i+1);\n");
    -- push value (it is of basic type or pointer)
    if TYPE[f.args[i].type] then
     write("    toLua_push"..TYPE[f.args[i].type].."("..f.args[i].name.."[i]);\n")
    elseif f.args[i].pointer == "*" then
     write("    toLua_pushusertag("..f.args[i].name.."[i],"..TAG[f.args[i].type]..");\n")
    else
     error("toLua - unexpected array type.")
    end
    write("    lua_storesubscript();\n");
    write("   }\n");
    write("   lua_endblock();\n");
    write("  }\n");
   end
   i = i+1
   j = j+1
  end
 end
 
 -- return value, if any
 if  f.type ~= "void" then
  if TYPE[f.type] then
   write("  toLua_push"..TYPE[f.type].."(L_result);\n")
  elseif f.pointer ~= "" then
   write("  toLua_pushusertag(")
   if f.pointer == "*" then
    write("L_result,")
   elseif f.pointer == "&" then
    write("&L_result,")
   else
    error("toLua - invalid function type.")
   end
   write(TAG[f.type]..");\n")
  else
   write("  toLua_pushuservalue((void*)&L_result,sizeof(L_result),")
   write(TAG[f.type]..");\n")
  end
 end
 write(" }\n")


 -- close function
 write("}\n\n")

end
 

-- write C code to bind index operator
function bindOperatorIndex (f,class)

 -- write get function
 write("static int "..f.cname.."_get (lua_Object lo1, lua_Object lo2)\n{\n") 

 -- find index type
 local indexType = ""
 if f.args[1].modifiers ~= "" then
  indexType = indexType..f.args[1].modifiers.." "
 end
 indexType = indexType..f.args[1].type

 -- declare "self" and "index" parameters
 write(" "..class.."* self;\n")
 write(" "..indexType.." index;\n")

 -- check and get "self" parameter
 write("\n")
 write(" if (!toLua_istype(lo1,"..TAG[class].."))")
 write("  return 0;\n")
 write(" else")
 write("  self = ("..class.."*)toLua_getuserdata(lo1);\n")
 
 -- check and get "index" value
 write("\n")
 write(" if (!toLua_istype(lo2,"..TAG[f.args[1].type].."))")
 write("  return 0;\n")
 write(" else")
 write("  index = ("..indexType..")")
 if TYPE[f.args[1].type] then
  write("toLua_get"..TYPE[f.args[1].type].."(lo2);")
 else
  write("toLua_getusertag(lo2"..","..TAG[f.args[1].type]..");")
 end

 -- return indexed value to Lua
 if f.type ~= "void" then
  write("\n {\n  ")
  write(f.type..f.pointer.." L_result = ("..f.type..f.pointer..")")
  write("self->operator[](index);\n")
  if TYPE[f.type] then
   write("  toLua_push"..TYPE[f.type].."(L_result);\n")
  elseif f.pointer ~= "" then
   write("  toLua_pushusertag(")
   if f.pointer == "*" then
    write("L_result,")
   elseif f.pointer == "&" then
    write("&L_result,")
   else
    error("toLua - invalid function type.")
   end
   write(TAG[f.type]..");\n")
  else
   write("  toLua_pushuservalue((void*)&L_result,sizeof(L_result),")
   write(TAG[f.type]..");\n")
  end
 end
 write(" }\n")
 write(" return 1;\n")

 -- close function
 write("}\n\n")

 -- write set function, if applied
 if f.pointer == "&" then

  -- write get function
  write("static int "..f.cname.."_set ")
  write("(lua_Object lo1, lua_Object lo2, lua_Object lo3)\n{\n") 
 
  -- find value (function) type
  local valueType = ""
  if f.modifiers ~= "" then
   valueType = valueType..f.modifiers.." "
  end
  valueType = valueType..f.type

  -- declare "self", "index", and "value" parameters
  write(" "..class.."* self;\n")
  write(" "..indexType.." index;\n")
  write(" "..valueType.." value;\n")

  -- check and get "self" parameter
  write("\n")
  write(" if (!toLua_istype(lo1,"..TAG[class].."))")
  write("  return 0;\n")
  write(" else")
  write("  self = ("..class.."*)toLua_getuserdata(lo1);\n")
  
  -- check and get "index" value
  write("\n")
  write(" if (!toLua_istype(lo2,"..TAG[f.args[1].type].."))")
  write("  return 0;\n")
  write(" else")
  write("  index = ("..indexType..")")
  if TYPE[f.args[1].type] then
   write("toLua_get"..TYPE[f.args[1].type].."(lo2);")
  else
   write("toLua_getusertag(lo2"..","..TAG[f.args[1].type]..");")
  end
 
  -- check and get "value" value
  write("\n")
  write(" if (!toLua_istype(lo3,"..TAG[f.type].."))")
  write("  return 0;\n")
  write(" else")
  write("  value = ("..valueType..")")
  if TYPE[f.type] then
   write("toLua_get"..TYPE[f.type].."(lo3)")
  else
   write("toLua_getusertag(lo3"..","..TAG[f.type]..")")
  end
  write(";\n\n")
 
  -- assign value
  write(" self->operator[](index) = value;\n");
  write(" return 1;\n")
 
  -- close function
  write("}\n\n")

 end
end
 

-- Write parameter list
function writeParamList (a)
 if a[1].type == "void" then
  return
 end
 local i=1
 while a[i] do
  if TYPE[a[i].type] or a[i].pointer == "*" then
   write(a[i].name)
  else
   write("*"..a[i].name)
  end
  i = i+1
  if a[i] then write(",") end
 end 
end


-- Bind classes
-- Write C++ code to bind classes
-- As input it receive the class list
function bindClasses (c)
 local i=1
 while c[i] do
  bindTypeTag(c[i].name)
  if c[i].base then
   bindTypeTag(c[i].base)
  end
  bindVariables(c[i].body.variables,c[i].name)
  bindFunctions(c[i].body.functions,c[i].name)

  -- write function to cast to base
  if c[i].base ~= "" then
   c[i].cast = "toLua_"..c[i].name.."_to_"..c[i].base
   write("static void* "..c[i].cast.." (void* derived)\n")
   write("{\n")
   write(" "..c[i].base.."* base = ("..c[i].name.."*)derived;\n")
   write(" return (void*)base;\n")
   write("}\n\n")
  end

  -- write init functin
  write("static void toLua_init_"..c[i].name.."(void)\n")
  write("{\n")
  -- create class table
  write(" lua_pushusertag(NULL,"..TAG[c[i].name]..");")
  write(" lua_storeglobal(\""..c[i].name.."\");\n")
  -- register parent
  if c[i].base ~= "" then
   write(" toLua_registerparent("..TAG[c[i].name]..","..
         TAG[c[i].base]..","..c[i].cast..");\n"
        )
  end
  -- init constants
  local j=1
  while c[i].body.constants[j] do
   write(" lua_pushusertag(NULL,"..TAG[c[i].name]..");\n")
   write(" lua_pushstring(\""..c[i].body.constants[j].."\");")
   write(" toLua_pushnumber("..c[i].name.."::"..
         c[i].body.constants[j]..");\n"
        )
   write(" lua_storesubscript();\n")
   j = j+1
  end
  -- init variables
  local j=1
  while c[i].body.variables[j] do
   write(" "..c[i].body.variables[j].cname.."();\n")
   j = j+1
  end
  -- init functions
  local j=1
  while c[i].body.functions[j] do
   if not c[i].body.functions[j+1] or  
      c[i].body.functions[j].name ~= c[i].body.functions[j+1].name 
   then
    if not strfind(c[i].body.functions[j].cname,"operator_index") then
     write(" toLua_funcfb("..TAG[c[i].name]..",\""..
           c[i].body.functions[j].luaname.."\","..
           c[i].body.functions[j].cname..");\n"
          )
    else
     write(" toLua_getfb("..TAG[c[i].name]..",\""..
           c[i].body.functions[j].luaname.."_get\","..
           c[i].body.functions[j].cname.."_get);\n"
          )
     if c[i].body.functions[j].pointer == "&" then
      write(" toLua_setfb("..TAG[c[i].name]..",\""..
            c[i].body.functions[j].luaname.."_set\","..
            c[i].body.functions[j].cname.."_set);\n"
           )
     end
    end
   end
   j = j+1
  end
  write("}\n\n\n")
  i = i+1
 end
end


-- Write function to initialize tags
function writeInitTag ()
 write("static void toLua_init_tags  (void)\n")
 write("{\n")
 write(" toLua_libtag(\""..package.."\");\n")
 local i=1
 while USERTAG[i] do
  write(" "..USERTAG[i].tag.." = toLua_newtag(\""..USERTAG[i].type.."\");\n")
  i = i+1
 end
 write("}\n\n\n")
end

-- write head
function writeHead ()
 print("#ifdef __cplusplus")
 print("extern \"C\"{")
 print("#endif")
 print("#include \"lua.h\"")
 print("#include \"tolua.h\"")
 print("#ifdef __cplusplus")
 print("}")
 print("#endif")
 print("")
end

-- write tail
function writeTail ()
 -- initialize if C++
 print("")
 print("#ifdef __cplusplus")
 print("static int dummy = toLua_"..package.."_open();")
 print("#endif")
end

-- Bind chunk
-- Write C/C++ code to bind an entire chunk
-- As input it receives the chunk table
function bindChunk (t)

 writeHead()

 -- bind variables, functions, and classes
 bindVariables(t.variables)
 bindFunctions(t.functions)
 bindClasses(t.classes)

 -- write function to initialize tags
 writeInitTag()

 -- init chunk
 write("/* ----------------------------------- PUBLIC INTERFACE */\n")
 write("int toLua_"..package.."_open (void)\n")
 write("{\n")
 write(" lua_beginblock();\n")
 write(" toLua_init();\n")
 write(" toLua_init_tags();\n")

 -- bind constants
 write("\n /* Bind constants */\n")
 local i=1
 while t.constants[i] do
  write(" toLua_pushnumber("..t.constants[i]..");")
  write(" lua_storeglobal(\""..t.constants[i].."\");\n")
  i = i+1
 end

 -- init variables
 write("\n /* Bind variables */\n")
 local i=1
 while t.variables[i] do
  write(" "..t.variables[i].cname.."();\n")
  i = i+1
 end

 -- init functions
 write("\n /* Bind functions */\n")
 local i=1
 while t.functions[i] do
  if not t.functions[i+1] or
     t.functions[i].name ~= t.functions[i+1].name
  then
   write(" lua_register(\""..t.functions[i].luaname.."\","..
         t.functions[i].cname..");\n"
        )
  end
  i = i+1;
 end

 -- init classes
 write("\n /* Bind classes */\n")
 local i=1
 while t.classes[i] do
  write(" toLua_init_"..t.classes[i].name.."();\n")
  i = i+1
 end
 write(" lua_endblock();\n")
 write(" return 1;\n")
 write("}\n")

 writeTail()
 
end

-- BIND PACKAGE
if not package then
 error("undefined 'package' variable for tolua.")
end

if not readfrom(package..".pkg") then
 error ("Could not find file "..package..".pkg")
end
s = read(".*")

-- eliminate comments
s = gsub(s,"//[^\n]*","")
c1 = format('%c', 2)
c2 = format('%c', 3)
s = gsub(s,"/%*",c1)
s = gsub(s,"%*/",c2)
s = gsub(s,c1.."[^"..c2.."]*"..c2," ")

-- open parser file
parserfile = getenv("TOLUA")
if parserfile then
 if strsub(parserfile,strlen(parserfile)) ~= "/" then
  parserfile = parserfile .. "/"
 end
 parserfile = parserfile .. "cparser.lua"
else
 parserfile = "cparser.lua"
end
if not dofile(parserfile) then
 error("Could not find 'cparser.lua' file.\n"..
       "You can set the environment variable 'TOLUA' to specify its directory."
      )
end

-- mark end of class
s = gsub(s,"\n}","\n}"..ENDCLASS)

t = parseChunk(s)
bindChunk(t)
--printChunk("MAIN",t)
